<?php

namespace Drupal\openstack_provider\Form\Config;

use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Render\AttachmentsInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Ajax\AjaxResponse;

/**
 * OpenStack launch form configurations.
 */
class OpenStackProviderAdminSettingsForm extends ConfigFormBase {

  /**
   * Drupal\Core\Config\ConfigFactoryInterface definition.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Drupal\cloud\Service\CloudServiceInterface definition.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloud;

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * OpenStackProviderAdminSettingsForm constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud
   *   The cloud service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The Messenger Service.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   *   The entity type manager object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    CloudServiceInterface $cloud,
    MessengerInterface $messenger,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    AccountInterface $current_user,
    EntityTypeManager $entity_type_manager
  ) {
    parent::__construct($config_factory);
    $this->configFactory = $config_factory;
    $this->cloud = $cloud;
    $this->messenger = $messenger;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->currentUser = $current_user;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('cloud'),
      $container->get('messenger'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('current_user'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return [
      'openstack_provider.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'openstack_provider_admin_settings';
  }

  /**
   * Ajax callback function to update vpc, and subnets.
   *
   * This method is called when the target provider option changes.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Returns the data back as AjaxResponse object.
   */
  public function vpcCallback(array $form, FormStateInterface $form_state): AttachmentsInterface {
    $target_provider = $form_state->getValue('target_provider');

    $vpc = $form['aws']['network']['vpc'];
    $vpc['#options'] = $this->getVpcs($target_provider, TRUE);

    $public_subnet = $form['aws']['network']['public_subnet_1'];
    $public_subnet['#options'] = $this->getSubnets($target_provider, TRUE);

    $private_subnet_1 = $form['aws']['network']['private_subnet_1'];
    $private_subnet_1['#options'] = $this->getSubnets($target_provider, TRUE);

    $private_subnet_2 = $form['aws']['network']['private_subnet_2'];
    $private_subnet_2['#options'] = $this->getSubnets($target_provider, TRUE);

    $response = new AjaxResponse();
    $response->addCommand(
      new ReplaceCommand('#vpc-wrapper', $vpc)
    );
    $response->addCommand(
      new ReplaceCommand('#public-subnet-wrapper', $public_subnet)
    );
    $response->addCommand(
      new ReplaceCommand('#private-subnet-1-wrapper', $private_subnet_1)
    );
    $response->addCommand(
      new ReplaceCommand('#private-subnet-2-wrapper', $private_subnet_2)
    );
    return $response;
  }

  /**
   * Ajax callback function to retrieve subnets.
   *
   * This method is called when the VPC option is changed.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Returns the data back as AjaxResponse object.
   */
  public function subnetCallback(array $form, FormStateInterface $form_state): AttachmentsInterface {
    $target_provider = $form_state->getValue('target_provider');
    $user_input = $form_state->getUserInput();
    $vpc = $user_input['vpc'] ?? '';

    $public_subnet = $form['aws']['network']['public_subnet_1'];
    $public_subnet['#options'] = $this->getSubnets($target_provider, $vpc, TRUE);

    $private_subnet_1 = $form['aws']['network']['private_subnet_1'];
    $private_subnet_1['#options'] = $this->getSubnets($target_provider, $vpc, TRUE);

    $private_subnet_2 = $form['aws']['network']['private_subnet_2'];
    $private_subnet_2['#options'] = $this->getSubnets($target_provider, $vpc, TRUE);

    $response = new AjaxResponse();

    $response->addCommand(
      new ReplaceCommand('#public-subnet-wrapper', $public_subnet)
    );
    $response->addCommand(
      new ReplaceCommand('#private-subnet-1-wrapper', $private_subnet_1)
    );
    $response->addCommand(
      new ReplaceCommand('#private-subnet-2-wrapper', $private_subnet_2)
    );
    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $config = $this->config('openstack_provider.settings');

    $form['openstack'] = [
      '#type' => 'details',
      '#title' => $this->t('OpenStack'),
      '#open' => TRUE,
    ];

    $providers = $this->getProviders('openstack');
    $form['openstack']['selectable_regions'] = [
      '#type' => 'select',
      '#title' => $this->t('Selectable OpenStack regions'),
      '#description' => $this->t('Regions available to users during tenant creation.'),
      '#options' => $providers,
      '#default_value' => $config->get('selectable_regions'),
      '#multiple' => TRUE,
      '#size' => count($providers),
    ];

    $form['aws'] = [
      '#type' => 'details',
      '#title' => $this->t('AWS Settings'),
      '#open' => TRUE,
    ];
    $form['aws']['target_provider'] = [
      '#type' => 'select',
      '#title' => $this->t('Target provider'),
      '#required' => TRUE,
      '#options' => $this->getProviders('aws_cloud', 'getCloudContext'),

      '#description' => $this->t('Select region to launch Cloud Orchestrator instances.'),
      '#ajax' => [
        'callback' => '::vpcCallback',
        'event' => 'change',
      ],
    ];

    if (!empty($config->get('target_provider'))) {
      $form['aws']['target_provider']['#default_value'] = $config->get('target_provider');
    }

    $form['aws']['network'] = [
      '#type' => 'details',
      '#title' => $this->t('Network'),
      '#open' => TRUE,
    ];

    $form['aws']['network']['vpc'] = [
      '#type' => 'select',
      '#title' => $this->t('VPC'),
      '#description' => $this->t('Select a VPC to use with this stack.'),
      '#required' => TRUE,
      '#options' => $this->getVpcs($config->get('target_provider') ?? ''),
      '#prefix' => '<div id="vpc-wrapper">',
      '#suffix' => '</div>',
      '#value' => $this->getSelectedOption('vpc', $form_state),
      '#ajax' => [
        'callback' => '::subnetCallback',
        'event' => 'change',
      ],
    ];

    $vpc_id = !empty($form['aws']['network']['vpc']['#value']) ? $form['aws']['network']['vpc']['#value'] : '';

    $form['aws']['network']['public_subnet_1'] = [
      '#type' => 'select',
      '#title' => $this->t('Public subnet'),
      '#description' => $this->t('Select a public subnet to use with the EC2 instance.'),
      '#options' => $this->getSubnets($config->get('target_provider') ?? '', $vpc_id),
      '#prefix' => '<div id="public-subnet-wrapper">',
      '#suffix' => '</div>',
      '#required' => TRUE,
      '#value' => $this->getSelectedOption('public_subnet_1', $form_state),
    ];

    $form['aws']['network']['private_subnet_1'] = [
      '#type' => 'select',
      '#title' => $this->t('Private subnet 1'),
      '#description' => $this->t('Select a private subnet for RDS, security groups and cache
      resources. Private Subnet 1 and Private Subnet 2 must be in two different
      Availability Zones in the same region.'),
      '#options' => $this->getSubnets($config->get('target_provider') ?? '', $vpc_id),
      '#prefix' => '<div id="private-subnet-1-wrapper">',
      '#suffix' => '</div>',
      '#required' => TRUE,
      '#value' => $this->getSelectedOption('private_subnet_1', $form_state),
    ];

    $form['aws']['network']['private_subnet_2'] = [
      '#type' => 'select',
      '#title' => $this->t('Private subnet 2'),
      '#description' => $this->t('Select a private subnet for RDS, security groups and cache
      resources. Private Subnet 1 and Private Subnet 2 must be in two different
      Availability Zones in the same region.'),
      '#options' => $this->getSubnets($config->get('target_provider') ?? '', $vpc_id),
      '#prefix' => '<div id="private-subnet-2-wrapper">',
      '#suffix' => '</div>',
      '#required' => TRUE,
      '#value' => $this->getSelectedOption('private_subnet_2', $form_state),
    ];

    $form['cfn'] = [
      '#type' => 'details',
      '#title' => $this->t('CloudFormation'),
      '#open' => TRUE,
    ];

    $form['cfn']['instance_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Instance type'),
      '#description' => $this->t('The EC2 instance type'),
      '#default_value' => $config->get('instance_type'),
      '#required' => TRUE,
    ];

    $form['cfn']['cloud_orchestrator_version'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Cloud Orchestrator version'),
      '#description' => $this->t('Version of the Cloud Orchestrator to install.'),
      '#default_value' => $config->get('cloud_orchestrator_version'),
      '#required' => TRUE,
    ];

    $form['cfn']['creation_timeout'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Creation timeout'),
      '#description' => $this->t('The length of time that CloudFormation waits until template
      creation is considered unsuccessful and is deleted.'),
      '#default_value' => $config->get('creation_timeout'),
      '#required' => TRUE,
    ];

    $form['cfn']['docker_file_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Docker file url'),
      '#description' => $this->t('The Dockerfile used to build Cloud Orchestrator container'),
      '#default_value' => $config->get('docker_file_url'),
      '#required' => TRUE,
    ];

    $form['rds'] = [
      '#type' => 'details',
      '#title' => $this->t('RDS'),
      '#open' => TRUE,
    ];

    $form['rds']['rds_instance_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('RDS instance type'),
      '#default_value' => $config->get('rds_instance_type'),
      '#required' => TRUE,
    ];

    $form['rds']['mysql_username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('MySQL username'),
      '#default_value' => $config->get('mysql_username'),
      '#required' => TRUE,
    ];

    $form['rds']['database_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Database name'),
      '#default_value' => $config->get('database_name'),
      '#required' => TRUE,
    ];

    $form['rds']['db_allocated_storage'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Database size'),
      '#description' => $this->t('The size of the database (Gb).  Must be between 5 and 16384.'),
      '#default_value' => $config->get('db_allocated_storage'),
      '#required' => TRUE,
    ];

    $form['drupal'] = [
      '#type' => 'details',
      '#title' => $this->t('Drupal'),
      '#open' => TRUE,
    ];

    $form['drupal']['drupal_site_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Site name'),
      '#default_value' => $config->get('drupal_site_name'),
      '#description' => $this->t('Site name for Cloud Orchestrator instance.'),
      '#required' => TRUE,
    ];

    $form['drupal']['drupal_site_email'] = [
      '#type' => 'email',
      '#title' => $this->t('Site email'),
      '#description' => $this->t('The From address in automated emails sent during registration and new password requests, and other notifications.'),
      '#default_value' => $config->get('drupal_site_email'),
      '#required' => TRUE,
    ];

    $zones = system_time_zones(NULL, TRUE);

    $form['drupal']['drupal_timezone'] = [
      '#type' => 'select',
      '#options' => $zones,
      '#title' => $this->t('Timezone'),
      '#default_value' => $config->get('drupal_timezone'),
      '#description' => $this->t('Default timezone.'),
      '#required' => TRUE,
    ];
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $config = $this->configFactory()->getEditable('openstack_provider.settings');
    $form_state->cleanValues();
    // Since some values are coming from Ajax, getUserInput is more
    // reliable to retrieve user input.
    foreach ($form_state->getUserInput() ?: [] as $key => $value) {
      if ($key === 'selectable_regions') {
        // This is a multiselect field.
        $config->set($key, $value);
      }
      else {
        $config->set($key, Html::escape($value));
      }
    }
    $config->save();
    parent::submitForm($form, $form_state);
  }

  /**
   * Get a list of OpenStack cloud providers.
   *
   * @return array
   *   An array list of providers.
   */
  private function getOpenStackProviders(): array {
    $providers = [];
    $entities = $this->cloudConfigPluginManager->loadConfigEntities('openstack');
    foreach ($entities ?: [] as $entity) {
      if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
        $providers[$entity->getCloudContext()] = $entity->label();
      }
    }
    return $providers;
  }

  /**
   * Helper method to get the user selected value from a dropdown element.
   *
   * @param string $form_key
   *   Key to search for.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @return string|null
   *   Selected value.
   */
  private function getSelectedOption(string $form_key, FormStateInterface $form_state): ?string {
    $config = $this->config('openstack_provider.settings');
    $option = $form_state->getValue($form_key) ?? $config->get($form_key);
    $user_input = $form_state->getUserInput();
    if (isset($user_input[$form_key])) {
      $option = $user_input[$form_key];
    }
    return $option;
  }

  /**
   * Get a list of cloud providers based on entity_bundle.
   *
   * @param string $entity_bundle
   *   The bundle to retrieve.
   * @param string $method_name
   *   The method name to act as the key.
   *
   * @return array
   *   An array list of providers.
   */
  private function getProviders(string $entity_bundle, string $method_name = 'id'): array {
    $providers = [];
    if (empty($entity_bundle)) {
      return [];
    }
    $entities = $this->cloudConfigPluginManager->loadConfigEntities($entity_bundle);
    foreach ($entities ?: [] as $entity) {
      if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
        $providers[$entity->$method_name()] = $entity->label();
      }
    }
    return $providers;
  }

  /**
   * Get vpcs depending on target provider.
   *
   * @param string $target_provider
   *   The target provider.
   * @param bool $add_none
   *   TRUE to add default empty option; needed for AJAX form updates.
   *
   * @return array
   *   An array of vpcs formatted for a select box.
   */
  private function getVpcs(string $target_provider, bool $add_none = FALSE): array {
    $vpcs = [];
    try {
      $entities = $this->entityTypeManager
        ->getStorage('aws_cloud_vpc')
        ->loadByProperties(['cloud_context' => $target_provider]);
      if ($add_none === TRUE) {
        $vpcs['_none'] = t('- Select -');
      }
      foreach ($entities ?: [] as $entity) {
        if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
          $vpcs[$entity->getVpcId()] = $this->t('%label', [
            '%label' => $entity->label(),
          ]);
        }
      }

    }
    catch (\Exception $e) {
      \Drupal::service('cloud')->handleException($e);
      $vpcs = [];
    }
    return $vpcs;
  }

  /**
   * Get subnets depending on target provider.
   *
   * @param string $target_provider
   *   The target provider.
   * @param string $vpc_id
   *   The vpc_id to query from.
   * @param bool $add_none
   *   TRUE to add default empty option; needed for AJAX form updates.
   *
   * @return array
   *   An array of subnets formatted for a select box.
   */
  private function getSubnets(string $target_provider, string $vpc_id, bool $add_none = FALSE): array {
    $subnets = [];
    try {
      $entities = $this->entityTypeManager
        ->getStorage('aws_cloud_subnet')
        ->loadByProperties([
          'cloud_context' => $target_provider,
          'vpc_id' => $vpc_id,
        ]);
      if ($add_none === TRUE) {
        $subnets['_none'] = t('- Select -');
      }
      foreach ($entities ?: [] as $entity) {
        if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
          $subnets[$entity->getSubnetId()] = $this->t('%label - (%cidr)', [
            '%label' => $entity->label(),
            '%cidr' => $entity->getCidrBlock(),
          ]);
        }
      }
    }
    catch (\Exception $e) {
      \Drupal::service('cloud')->handleException($e);
      $subnet = [];
    }
    return $subnets;
  }

}
