<?php

namespace Drupal\openstack_provider\Service;

use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\cloud\Entity\CloudConfig;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\Yaml;
use Drupal\Component\Uuid\Php;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ExtensionPathResolver;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Password\DefaultPasswordGenerator;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\facade\Entity\TenantInterface;
use Drupal\openstack\Service\OpenStackServiceFactoryInterface;
use Drupal\openstack_provider\Entity\OpenStackTenantProjectInterface;
use Drupal\openstack_provider\Entity\OpenStackTenantProject;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Service providing generic methods for openstack_provider.
 */
class OpenStackProviderService implements OpenStackProviderServiceInterface, ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * Description string to send to OpenStack.
   */
  public const DESCRIPTION_STRING = 'auto-created-from-facade';

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The config factory.
   *
   * Subclasses should use the self::config() method, which may be overridden to
   * address specific needs when loading config, rather than this property
   * directly. See \Drupal\Core\Form\ConfigFormBase::config() for an example of
   * this.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The path resolver.
   *
   * @var \Drupal\Core\Extension\ExtensionPathResolver
   */
  protected $extensionPathResolver;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * The password generator service.
   *
   * @var \Drupal\Core\Password\DefaultPasswordGenerator
   */
  private $passwordGenerator;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The OpenStack Service Factory.
   *
   * @var \Drupal\openstack\Service\OpenStackServiceFactoryInterface
   */
  private $openStackServiceFactory;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  private $currentUser;

  /**
   * The datetime interface.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  private $dateTime;

  /**
   * The uuid service.
   *
   * @var \Drupal\Component\Uuid\Php
   */
  private $uuid;

  /**
   * Guzzle client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager interface.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Configuration factory.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   File system interface.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   Module handler interface.
   * @param \Drupal\Core\Extension\ExtensionPathResolver $extension_path_resolver
   *   Extension path resolver.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger object.
   * @param \Drupal\openstack\Service\OpenStackServiceFactoryInterface $openstack_service_factory_interface
   *   The OpenStack factory interface.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The Ec2 service.
   * @param \Drupal\Core\Password\DefaultPasswordGenerator $password_generator
   *   The password generator.
   * @param \Drupal\Component\Datetime\TimeInterface $date_time
   *   The time object.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user.
   * @param \Drupal\Component\Uuid\Php $uuid
   *   The uuid service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The Guzzle Http client.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    FileSystemInterface $file_system,
    ModuleHandlerInterface $module_handler,
    ExtensionPathResolver $extension_path_resolver,
    Messenger $messenger,
    OpenStackServiceFactoryInterface $openstack_service_factory_interface,
    Ec2ServiceInterface $ec2_service,
    DefaultPasswordGenerator $password_generator,
    TimeInterface $date_time,
    AccountProxyInterface $current_user,
    Php $uuid,
    ClientInterface $http_client
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->configFactory = $config_factory;
    $this->fileSystem = $file_system;
    $this->moduleHandler = $module_handler;
    $this->extensionPathResolver = $extension_path_resolver;
    $this->messenger = $messenger;
    $this->openStackServiceFactory = $openstack_service_factory_interface;
    $this->ec2Service = $ec2_service;
    $this->passwordGenerator = $password_generator;
    $this->currentUser = $current_user;
    $this->dateTime = $date_time;
    $this->uuid = $uuid;
    $this->httpClient = $http_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('config.factory'),
      $container->get('file_system'),
      $container->get('module_handler'),
      $container->get('extension.path.resolver'),
      $container->get('messenger'),
      $container->get('openstack.factory'),
      $container->get('openstack.ec2'),
      $container->get('password_generator'),
      $container->get('datetime.time'),
      $container->get('current_user'),
      $container->get('uuid'),
      $container->get('http_client')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function compileYamlParameters(string $yaml, EntityInterface $entity): array {

    $yaml = $this->decodeYaml($yaml);
    $parsed_parameters = [];
    if (empty($yaml['parameters'])) {
      $this->messenger->addError($this->t('Unable to retrieve parameters.'));
      return [];
    }
    foreach ($yaml['parameters'] ?: [] as $key => $parameter) {
      switch ($parameter['type']) {
        case 'config':
          $parsed_parameters[$key] = $this->getConfigurationValue($parameter['config_name'], $parameter['config_key']);
          break;

        case 'auto':
          $method = $parameter['method'];
          if (!method_exists($this, $method)) {
            break;
          }
          $parsed_parameters[$key] = $this->$method($entity);
          break;

        case 'entity':
          $parsed_parameters[$key] = $this->getEntityValue($entity, $parameter['entity_type'], $parameter['entity_bundle'], $parameter['field']);
          break;

        default:
          break;
      }
    }
    // @todo Check if any parameters are empty and halt the launch.
    return [
      'cfn_body' => file_get_contents($this->getRealPath($yaml['cloud_formation_template'])),
      'parameters' => $parsed_parameters,
      'target_provider' => $this->getConfigurationValue('openstack_provider.settings', 'target_provider'),
    ];

  }

  /**
   * {@inheritdoc}
   */
  public function createRemoteCloudConfig(TenantInterface $tenant, array $params): void {
    $headers = [
      'Authorization' => 'Bearer ' . base64_encode($tenant->get('field_bearer_token')->value ?? ''),
      'Content-Type' => 'application/json',
    ];
    /** @var \Drupal\aws_cloud\Entity\CloudFormation\Stack $stack */
    $stack = $tenant->get('field_aws_cloud_stack')->entity;
    if (empty($stack)) {
      throw new \Exception('Cannot add project. No stack found.');
    }
    $output = $stack->getOutputs();
    $array_index = array_search('DrupalUrl', array_column($output, 'item_key'), TRUE);
    $endpoint = $output[(int) $array_index]['item_value'];
    if (empty($endpoint)) {
      throw new \Exception('Cannot add project. No endpoint found.');
    }
    // Need error handling of endpoint.
    $response = $this->httpClient->post(
      $endpoint . '/entity/cloud_config',
      [
        'body' => json_encode($params, JSON_THROW_ON_ERROR),
        'headers' => $headers,
      ]
    );
    if ($response->getStatusCode() !== 201) {
      throw new \Exception("Error adding project to $endpoint");
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackResource(CloudConfig $cloud_config, TenantInterface $tenant, string $resource_prefix, string $project_entity_name): OpenStackTenantProjectInterface {
    $ec2_service = $this->openStackServiceFactory->get($cloud_config->getCloudContext());
    $this->ec2Service = $ec2_service;

    // Fetch the domain_id.
    $domain_id = $cloud_config->get('field_domain_id')->value;

    // Setup some defaults values to pass to the APIs.
    $project_name = $this->generateNameString($resource_prefix, 'project');
    $username = $this->generateNameString($resource_prefix, 'user');
    $role_name = $this->generateNameString($resource_prefix, 'role');
    $password = $this->passwordGenerator->generate();

    $project_id = $this->createOpenStackProject($domain_id, $project_name);
    $role_id = $this->createOpenStackRole($role_name);
    $user_id = $this->createOpenStackUser($username, $password, $tenant->get('field_email')->value);
    $this->assignRoleToUserOnProject($project_id, $user_id, $role_id);

    // Add an OpenStackTenantProject entity with all the generated info.
    $tenant_project_entity = OpenStackTenantProject::create([
      'name' => $project_entity_name,
      'created' => $this->dateTime->getRequestTime(),
      'uid' => $this->currentUser->id(),
      'tenant' => $tenant->id(),
      'openstack_region' => $cloud_config->id(),
      'username' => $username,
      'password' => $password,
      'project_id' => $project_id,
      'domain_id' => $domain_id,
      'role_id' => $role_id,
      'user_id' => $user_id,
      'project_name' => $project_name,
      'role_name' => $role_name,
    ]);
    $tenant_project_entity->save();
    return $tenant_project_entity;
  }

  /**
   * Generate a unique string for OpenStack api calls.
   *
   * @param string $prefix
   *   String prefix.
   * @param string $type
   *   Entity type.
   *
   * @return string
   *   Generated string.
   */
  private function generateNameString(string $prefix, string $type): string {
    return $this->t('@prefix-@type-@unique', [
      '@prefix' => $this->formatName($prefix),
      '@type' => $type,
      '@unique' => $this->generateUniqueSuffix(),
    ])->render();
  }

  /**
   * Helper to generate unique suffix.
   *
   * Currently, timestamp is used as the unique identified, but this can change
   * in the future.
   *
   * @return string
   *   The unique identifier.
   */
  private function generateUniqueSuffix(): string {
    return time();
  }

  /**
   * Helper to create the OpenStack project.
   *
   * @param string $domain_id
   *   The domain_id.
   * @param string $project_name
   *   The project name.
   *
   * @return string|null
   *   OpenStack project id or null if there was an error.
   */
  private function createOpenStackProject(string $domain_id, string $project_name): ?string {
    $project = $this->ec2Service->createProject([
      'DomainId' => $domain_id,
      'Name' => $project_name,
      'Description' => self::DESCRIPTION_STRING,
      'Enabled' => TRUE,
    ]);
    if (empty($project)) {
      throw new Exception('Error creating OpenStack project.');
    }
    return $project['Project']['Id'];
  }

  /**
   * Helper to create the OpenStack role.
   *
   * @param string $role_name
   *   The role name.
   *
   * @return string|null
   *   OpenStack role id or null if there was an error.
   */
  private function createOpenStackRole(string $role_name): ?string {
    $role = $this->ec2Service->createRole([
      'Name' => $role_name,
      'Description' => self::DESCRIPTION_STRING,
    ]);
    if (empty($role)) {
      throw new Exception('Error creating OpenStack role.');
    }
    return $role['Role']['Id'];
  }

  /**
   * Helper to create the OpenStack user.
   *
   * @param string $user_name
   *   The username.
   * @param string $password
   *   Password string.
   * @param string $email
   *   The email string.
   *
   * @return string|null
   *   OpenStack role id or null if there was an error.
   */
  private function createOpenStackUser(string $user_name, string $password, string $email): ?string {
    $user = $this->ec2Service->createUser([
      'Description' => self::DESCRIPTION_STRING,
      'Name' => $user_name,
      'Password' => $password,
      'Email' => $email,
    ]);
    if (empty($user)) {
      throw new Exception('Error creating OpenStack user.');
    }
    return $user['User']['Id'];
  }

  /**
   * Assign a user to a project.
   *
   * @param string $project_id
   *   The project id.
   * @param string $user_id
   *   The user id.
   * @param string $role_id
   *   The role id.
   */
  private function assignRoleToUserOnProject(string $project_id, string $user_id, string $role_id): void {
    $this->ec2Service->assignRoleToUserOnProject([
      'ProjectId' => $project_id,
      'RoleId' => $role_id,
      'UserId' => $user_id,
    ]);
  }

  /**
   * Generate the CFn stack name.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Tenant entity.
   *
   * @return string
   *   The generated stack name.
   */
  private function generateStackName(EntityInterface $entity): string {
    return $this->t(':name-:timestamp', [
      ':name' => $this->formatName($entity->label()),
      ':timestamp' => time(),
    ])->render();
  }

  /**
   * Generate the CFn stack prefix.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Tenant entity.
   *
   * @return string
   *   The generated stack name.
   */
  private function generateStackPrefix(EntityInterface $entity): string {
    // Stack prefix can only be 20 characters in length.
    return $this->t(':name-:timestamp', [
      ':name' => substr($this->formatName($entity->label()), 0, 9),
      ':timestamp' => time(),
    ])->render();
  }

  /**
   * Helper function to strip tenant entity name.
   *
   * The formatted tenant name is used for CloudFormnation stack name and
   * can only contain alphanumeric characters and hyphens.
   *
   * @param string $name
   *   Name to format.
   *
   * @return string
   *   Formatted name.
   */
  private function formatName(string $name): string {
    $name = preg_replace('/[^a-zA-Z0-9]/', '', $name);
    return strtolower($name);
  }

  /**
   * Generate bearer token.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Tenant entity.
   *
   * @return string
   *   The generated bearer token.
   */
  private function generateBearerToken(EntityInterface $entity): string {
    return $this->uuid->generate();
  }

  /**
   * Get entity value given an entity field name.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity to retrieve from.
   * @param string $entity_type
   *   The entity type.
   * @param string $entity_bundle
   *   The entity bundle.
   * @param string $field
   *   Field to retrieve.
   *
   * @return string|null
   *   field value or null.
   */
  private function getEntityValue(EntityInterface $entity, string $entity_type, string $entity_bundle, string $field): ?string {
    if ($entity->getEntityTypeId() !== $entity_type && $entity->bundle() !== $entity_bundle) {
      return NULL;
    }
    return $entity->get($field)->value ?? NULL;
  }

  /**
   * Helper method to retrieve value from a specific configuration.
   *
   * @param string $config_name
   *   Configuration name.
   * @param string $config_key
   *   Configuration key.
   *
   * @return string|null
   *   Configuration value or null.
   */
  private function getConfigurationValue(string $config_name, string $config_key): ?string {
    return $this->configFactory->get($config_name)
      ->get($config_key);
  }

  /**
   * Given relative path of yaml file, resolve to absolute file path.
   *
   * @param string $path
   *   Path relative to openstack_provider module.
   *
   * @return string
   *   Absolute file path.
   */
  private function getRealPath($path): string {
    return realpath($this->extensionPathResolver->getPath('module', 'openstack_provider')) . '/' . $path;
  }

  /**
   * Decode the yaml file.
   *
   * @param string $path
   *   Yaml file to decode.
   *
   * @return array
   *   Full path to yaml file.
   */
  private function decodeYaml(string $path): array {
    $path = $this->getRealPath($path);
    return Yaml::decode(file_get_contents($path));
  }

}
